from abc import ABC, abstractmethod
from typing import Optional

from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.schedule_spec import WorkSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator


class Timeline(ABC):
    """
    Entity that saves info on the use of resources over time.
    Timeline provides opportunities to work with GraphNodes and resources over time.
    """

    @abstractmethod
    def schedule(self,
                 node: GraphNode,
                 node2swork: dict[GraphNode, ScheduledWork],
                 passed_agents: list[Worker],
                 contractor: Contractor,
                 spec: WorkSpec,
                 assigned_start_time: Time | None = None,
                 assigned_time: Time | None = None,
                 assigned_parent_time: Time = Time(0),
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) -> Time:
        """
        Schedules the given `GraphNode` using passed agents, spec and times.
        If start time not passed, it should be computed as minimum work start time.
        :return: scheduled finish time of given work
        """
        ...

    def find_min_start_time(self,
                            node: GraphNode,
                            worker_team: list[Worker],
                            node2swork: dict[GraphNode, ScheduledWork],
                            spec: WorkSpec,
                            parent_time: Time = Time(0),
                            work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) -> Time:
        """
        Computes start time, max parent time, contractor and exec times for given node.

        :param worker_team: list of passed workers. It Should be IN THE SAME ORDER AS THE CORRESPONDING WREQS
        :param node: the GraphNode whose minimum time we are trying to find
        :param node2swork: dictionary, that match GraphNode to ScheduleWork respectively
        :param spec: specification for given `GraphNode`
        :param parent_time: the minimum start time
        :param work_estimator: function that calculates execution time of the GraphNode
        :return: minimum time
        """
        return self.find_min_start_time_with_additional(node, worker_team, node2swork, spec, None,
                                                        parent_time, work_estimator)[0]

    @abstractmethod
    def find_min_start_time_with_additional(self,
                                            node: GraphNode,
                                            worker_team: list[Worker],
                                            node2swork: dict[GraphNode, ScheduledWork],
                                            spec: WorkSpec,
                                            assigned_start_time: Optional[Time] = None,
                                            assigned_parent_time: Time = Time(0),
                                            work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
            -> tuple[Time, Time, dict[GraphNode, tuple[Time, Time]]]:
        ...

    @abstractmethod
    def can_schedule_at_the_moment(self,
                                   node: GraphNode,
                                   worker_team: list[Worker],
                                   spec: WorkSpec,
                                   node2swork: dict[GraphNode, ScheduledWork],
                                   start_time: Time,
                                   exec_time: Time) -> bool:
        """
        Returns the ability of scheduling given `node` at the `start_time` moment
        """
        ...

    @abstractmethod
    def update_timeline(self,
                        finish_time: Time,
                        exec_time: Time,
                        node: GraphNode,
                        worker_team: list[Worker],
                        spec: WorkSpec):
        ...



from random import Random

from sampo.generator import SyntheticGraphType
from sampo.generator.environment import get_contractor
from sampo.generator.pipeline.extension import extend_names, extend_resources
from sampo.generator.pipeline.project import get_small_graph, get_graph
from sampo.schemas.graph import WorkGraph


class SimpleSynthetic:
    """
    Designed to simplify the use of the synthetic data generator.
    """

    def __init__(self, rand: int | Random | None = None) -> None:
        if isinstance(rand, Random):
            self._rand = rand
        else:
            self._rand = Random(rand)

    def small_work_graph(self, cluster_name: str | None = 'C1') -> WorkGraph:
        """
        Creates a small graph of works consisting of 30-50 vertices;

        :param cluster_name: str - the first cluster name
        :return:
            work_graph: WorkGraph - work graph where count of vertex between 30 and 50
        """
        return get_small_graph(cluster_name, self._rand)

    def work_graph(self, mode: SyntheticGraphType | None = SyntheticGraphType.GENERAL,
                   cluster_counts: int | None = 0,
                   bottom_border: int | None = 0,
                   top_border: int | None = 0) -> WorkGraph:
        """
        Invokes a graph of the given type if at least one positive value of
            cluster_counts, bottom_border or top_border is given;

        :param mode: str - 'general' or 'sequence' or 'parallel - the type of the returned graph
        :param cluster_counts: Optional[int] - Number of clusters for the graph
        :param bottom_border: Optional[int] - bottom border for number of works for the graph
        :param top_border: Optional[int] - top border for number of works for the graph
        :return:
            work_graph: WorkGraph - the desired work graph
        """
        return get_graph(mode=mode, cluster_counts=cluster_counts, bottom_border=bottom_border, top_border=top_border,
                         rand=self._rand)

    def contractor(self, pack_worker_count: float):
        """
        Generates a contractor by pack_worker_count and from sampo.generator.environment.contractor.get_contractor
        with default optional parameters

        :param pack_worker_count: The number of resource sets
        :return: the contractor
        """
        return get_contractor(pack_worker_count, rand=self._rand)

    def advanced_work_graph(self, works_count_top_border: int, uniq_works: int, uniq_resources: int):
        """
        Invokes a graph of the given type for given works_count_top_border,
        expands the number of unique resources and job titles, if possible

        :param uniq_resources: Number of unique resources
        :param uniq_works: Number of unique work names
        :param works_count_top_border: Optional[int] - top border for number of works for the graph

        :return: the desired work graph
        """
        wg = self.work_graph(top_border=works_count_top_border)
        wg = extend_names(uniq_works, wg, self._rand)
        wg = extend_resources(uniq_resources, wg, self._rand)
        return wg



from copy import deepcopy
from dataclasses import dataclass
from typing import Callable

import numpy as np
import torch

from sampo.scheduler.base import Scheduler
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.multi_agency.block_graph import BlockGraph, BlockNode
from sampo.scheduler.multi_agency.exception import NoSufficientAgents
from sampo.scheduler.selection.neural_net import NeuralNetTrainer
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.utils.obstruction import Obstruction
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.time import Time


class Agent:
    """
    Agent entity representation in the multi-agent model
    Agent have 2 actions: give offer and accept offer
    """

    def __init__(self, name: str, scheduler: Scheduler, contractors: list[Contractor]):
        self.name = name
        self._timeline = None
        self._scheduler = scheduler
        self._contractors = contractors
        self._last_task_executed = Time(0)
        self._downtime = Time(0)

    def offer(self, wg: WorkGraph, parent_time: Time) -> tuple[Time, Time, Schedule, Timeline]:
        """
        Computes the offer from agent to manager. Handles all works from given wg.

        :param wg: the given block of tasks
        :param parent_time: max end time of parent blocks
        :return: offered start time, end time, resulting schedule and timeline before offering

        To apply returned offer, use `Agent#confirm`.
        """

        schedule, start_time, timeline, _ = \
            self._scheduler.schedule_with_cache(wg, self._contractors,
                                                assigned_parent_time=parent_time, timeline=deepcopy(self._timeline))
        return start_time, schedule.execution_time, schedule, timeline

    def confirm(self, timeline: Timeline, start: Time, end: Time):
        """
        Applies the given offer.

        :param timeline: timeline returned from corresponding `Agent#offer`
        :param start: global start time of confirmed block
        :param end: global end time of confirmed block
        """
        self._timeline = timeline
        self.update_stat(start)
        # update last task statistic
        self._last_task_executed = end

    def update_stat(self, start: Time):
        """
        Count last iteration downtime.
        If given start time is lower, than the last executed task, then this downtime are already in self_downtime

        :param start: global start time of confirmed block
        """
        self._downtime += max(Time(0), start - self._last_task_executed)

    def __str__(self):
        return f'Agent(name={self.name}, scheduler={self._scheduler}, downtime={self._downtime})'

    def __repr__(self):
        return str(self)

    @property
    def downtime(self) -> Time:
        return self._downtime

    @property
    def contractors(self):
        return self._contractors

    @property
    def end_time(self) -> Time:
        return self._last_task_executed

    @property
    def scheduler(self):
        return self._scheduler


@dataclass
class ScheduledBlock:
    """
    An object represents a scheduled graph block(group of works).

    Contains all data used in scheduling, the agent and resulting information.
    """
    wg: WorkGraph
    schedule: Schedule
    agent: Agent
    start_time: Time
    end_time: Time

    @property
    def id(self):
        return self.wg.start.id

    @property
    def duration(self):
        return self.end_time - self.start_time

    def __str__(self):
        return f'ScheduledBlock(start_time={self.start_time}, end_time={self.end_time}, agent={self.agent})'

    def __repr__(self):
        return str(self)


class Manager:
    """
    Manager entity representation in the multi-agent model
    Manager interact with agents

    :param agents: list of agents that has own scheduling algorithm and set of contractors
    """
    def __init__(self, agents: list[Agent]):
        if len(agents) == 0:
            raise NoSufficientAgents('Manager can not work with empty list of agents')
        self._agents = agents

    # TODO Upgrade to supply the best parallelism
    def manage_blocks(self, bg: BlockGraph, logger: Callable[[str], None] = None) -> dict[str, ScheduledBlock]:
        """
        Runs the multi-agent system based on auction on given BlockGraph.
        
        :param bg: 
        :param logger:
        :return: an index of resulting `ScheduledBlock`s built by ids of corresponding `WorkGraph`s
        """
        id2sblock = {}
        for i, block in enumerate(bg.toposort()):
            max_parent_time = max((id2sblock[parent.id].end_time for parent in block.blocks_from), default=Time(0)) + 1
            start_time, end_time, agent_schedule, agent \
                = self.run_auction_with_obstructions(block.wg, max_parent_time, block.obstruction)

            assert start_time >= max_parent_time, f'Scheduler {agent._scheduler} does not handle parent_time!'

            if logger and not block.is_service():
                logger(f'{agent._scheduler}')
            sblock = ScheduledBlock(wg=block.wg, agent=agent, schedule=agent_schedule,
                                    start_time=start_time,
                                    end_time=end_time)
            id2sblock[sblock.id] = sblock

        return id2sblock

    def run_auction_with_obstructions(self, wg: WorkGraph, parent_time: Time = Time(0),
                                      obstruction: Obstruction | None = None):
        if obstruction:
            obstruction.generate(wg)
        return self.run_auction(wg, parent_time)

    def run_auction(self, wg: WorkGraph, parent_time: Time = Time(0)) -> (Time, Time, Schedule, Agent):
        """
        Runs the auction on the given `WorkGraph`.

        :param wg: target `WorkGraph`
        :param parent_time: max parent time of given block
        :return: best start time, end time and the agent that is able to support this working time
        """
        best_start_time = 0
        best_end_time = Time.inf()
        best_schedule = None
        best_timeline = None
        best_agent = None

        offers = [(agent, agent.offer(wg, parent_time)) for agent in self._agents]

        for offered_agent, (offered_start_time, offered_end_time, offered_schedule, offered_timeline) in offers:
            if offered_end_time < best_end_time:
                best_start_time = offered_start_time
                best_end_time = offered_end_time
                best_schedule = offered_schedule
                best_timeline = offered_timeline
                best_agent = offered_agent
        best_agent.confirm(best_timeline, best_start_time, best_end_time)
        for agent in self._agents:
            if agent.name != best_agent.name:
                agent.update_stat(best_start_time)

        print(best_agent.name)
        return best_start_time, best_end_time, best_schedule, best_agent


class NeuralManager:
    """
    Manager entity representation in the multi-agent model, that uses neural networks
    Neural manager interact with agents
    Neural manager uses neural network as the method of the most suitable agent for each work graph

    :param agents: list of agents that has own scheduling algorithm and set of contractors
    :param algo_trainer: work with a neural network that predicts the most suitable scheduling algorithm
    :param contractor_trainer: work with a neural network that predicts a set of resources
    that the most suitable contractor should have
    :param algorithms: list of unique scheduling algorithms that agents have
    :param blocks: list of blocks of input Block Graph in topological order
    :param encoding_blocks: list of graph block embeddings
    """

    def __init__(self, agents: list[Agent],
                 algo_trainer: NeuralNetTrainer,
                 contractor_trainer: NeuralNetTrainer,
                 algorithms: list[GenericScheduler],
                 blocks: list[BlockNode],
                 encoding_blocks: list[torch.Tensor]):
        if len(agents) == 0:
            raise NoSufficientAgents('Manager can not work with empty list of agents')
        self._agents = agents
        self.algo_trainer = algo_trainer
        self.contractor_trainer = contractor_trainer
        self.algorithms = algorithms
        self.blocks = blocks
        self.encoding_blocks = encoding_blocks

    # TODO Upgrade to supply the best parallelism
    def manage_blocks(self, logger: Callable[[str], None] = None) -> dict[str, ScheduledBlock]:
        """
        Runs the multi-agent system based on auction on given BlockGraph.

        :param logger:
        :return: an index of resulting `ScheduledBlock`s built by ids of corresponding `WorkGraph`s
        """
        id2sblock = {}
        for i, block in enumerate(self.blocks):
            max_parent_time = max((id2sblock[parent.id].end_time for parent in block.blocks_from), default=Time(0)) + 1
            start_time, end_time, agent_schedule, agent \
                = self.run_auction_with_obstructions(block.wg, i, max_parent_time, block.obstruction)

            assert start_time >= max_parent_time, f'Scheduler {agent._scheduler} does not handle parent_time!'

            if logger and not block.is_service():
                logger(f'{agent._scheduler}')
            sblock = ScheduledBlock(wg=block.wg, agent=agent, schedule=agent_schedule,
                                    start_time=start_time,
                                    end_time=end_time)
            id2sblock[sblock.id] = sblock

        return id2sblock

    def run_auction_with_obstructions(self, wg: WorkGraph,
                                      index: int,
                                      parent_time: Time = Time(0),
                                      obstruction: Obstruction | None = None):
        if obstruction:
            obstruction.generate(wg)
        return self.run_auction(wg, index, parent_time)

    def run_auction(self, wg: WorkGraph, index: int, parent_time: Time = Time(0)) -> (Time, Time, Schedule, Agent):
        """
        Runs the auction on the given `WorkGraph`.

        :param index: index of agent from the list of agents
        :param wg: target `WorkGraph`
        :param parent_time: max parent time of given block
        :return: best start time, end time and the agent that is able to support this working time
        """

        wg_encoding = [self.encoding_blocks[index]]
        predicted = self.algo_trainer.predict(wg_encoding)
        predict_proba = self.algo_trainer.predict_proba(wg_encoding)
        best_algo = type(self.algorithms[int(predicted)])
        best_contractor = np.asarray(self.contractor_trainer.predict([self.encoding_blocks[index]]))[0]

        time_algo_agents = []
        not_time_algo_agents = []
        time_not_algo_agents = []
        not_time_not_algo_agents = []

        for agent in self._agents:
            if parent_time > agent.end_time and isinstance(agent.scheduler, best_algo):
                time_algo_agents.append(agent)
            elif not isinstance(agent.scheduler, best_algo) and parent_time > agent.end_time:
                time_not_algo_agents.append(agent)
            elif parent_time <= agent.end_time and isinstance(agent.scheduler, best_algo):
                not_time_algo_agents.append(agent)
            else:
                not_time_not_algo_agents.append(agent)

        def auction(agents: list[Agent]) -> Agent:
            less_mse = 10**9
            best_agent = None

            for agent in agents:
                resources = []
                for worker in agent.contractors[0].workers.values():
                    resources.append(worker.count)
                resources = np.asarray(resources)
                mse = sum((resources - best_contractor) ** 2)
                if less_mse > mse:
                    less_mse = mse
                    best_agent = agent

            return best_agent

        if len(time_algo_agents) > 0:
            best_agent = auction(time_algo_agents)
        elif len(time_not_algo_agents) > 0:
            best_agent = auction(time_not_algo_agents)
        elif len(not_time_algo_agents) > 0:
            best_agent = auction(not_time_algo_agents)
        else:
            best_agent = auction(not_time_not_algo_agents)

        best_start_time, best_end_time, best_schedule, best_timeline = best_agent.offer(wg, parent_time)

        best_agent.confirm(best_timeline, best_start_time, best_end_time)
        for agent in self._agents:
            if agent.name != best_agent.name:
                agent.update_stat(best_start_time)

        print(best_agent.name)
        return best_start_time, best_end_time, best_schedule, best_agent



from matplotlib.figure import Figure

from sampo.pipeline import PipelineError
from sampo.schemas import Schedule, WorkGraph, ScheduledProject
from sampo.utilities.visualization.base import VisualizationMode, visualize
from sampo.utilities.visualization.resources import resource_employment_fig, EmploymentFigType
from sampo.utilities.visualization.schedule import schedule_gant_chart_fig
from sampo.utilities.visualization.work_graph import work_graph_fig


class ScheduleVisualization:
    def __init__(self, schedule: Schedule, start_date: str):
        self._schedule = schedule.merged_stages_datetime_df(start_date)
        self._shape = 10, 10

    def fig(self, shape: tuple[int, int]) -> 'ScheduleVisualization':
        self._shape = shape
        return self

    def gant_chart(self, visualization_mode: VisualizationMode = VisualizationMode.ReturnFig) -> Figure | None:
        return schedule_gant_chart_fig(self._schedule, visualization=visualization_mode)

    def date_labeled_resource_chart(self, visualization_mode: VisualizationMode = VisualizationMode.ReturnFig) \
            -> Figure | None:
        return resource_employment_fig(self._schedule,
                                       fig_type=EmploymentFigType.DateLabeled,
                                       vis_mode=visualization_mode)

    def work_labeled_resource_chart(self, visualization_mode: VisualizationMode = VisualizationMode.ReturnFig) \
            -> Figure | None:
        return resource_employment_fig(self._schedule,
                                       fig_type=EmploymentFigType.WorkLabeled,
                                       vis_mode=visualization_mode)


class WorkGraphVisualization:
    def __init__(self, wg: WorkGraph):
        self._wg = wg
        self._shape = 10, 10

    def fig(self, shape: tuple[int, int]) -> 'WorkGraphVisualization':
        self._shape = shape
        return self

    def work_graph_chart(self, visualization_mode: VisualizationMode = VisualizationMode.ReturnFig) -> Figure | None:
        return visualize(work_graph_fig(self._wg, self._shape), visualization_mode)


class Visualization:

    def __init__(self):
        self.wg_vis = None
        self.schedule_vis = None

    @staticmethod
    def from_project(project: ScheduledProject, start_date: str) -> 'Visualization':
        vis = Visualization()
        vis.wg(project.wg)
        vis.schedule(project.schedule, start_date)
        return vis

    def wg(self, wg: WorkGraph):
        self.wg_vis = WorkGraphVisualization(wg)

    def schedule(self, schedule: Schedule, start_date: str):
        self.schedule_vis = ScheduleVisualization(schedule, start_date)

    def shape(self, fig_shape: tuple[int, int]) -> 'Visualization':
        self.wg_vis.fig(fig_shape)
        self.schedule_vis.fig(fig_shape)
        return self

    def show_gant_chart(self) -> 'Visualization':
        if self.schedule_vis is None:
            raise PipelineError('Schedule not specified')
        self.schedule_vis.gant_chart(VisualizationMode.ShowFig)
        return self

    def show_resource_charts(self) -> 'Visualization':
        if self.schedule_vis is None:
            raise PipelineError('Schedule not specified')
        self.schedule_vis.date_labeled_resource_chart(VisualizationMode.ShowFig)
        self.schedule_vis.work_labeled_resource_chart(VisualizationMode.ShowFig)
        return self

    def show_work_graph(self) -> 'Visualization':
        if self.wg_vis is None:
            raise PipelineError('WorkGraph not specified')
        self.wg_vis.work_graph_chart(VisualizationMode.ShowFig)
        return self

    def get_all_figs(self):
        return {
            'date_labeled_resource_chart': self.schedule_vis.date_labeled_resource_chart(VisualizationMode.ReturnFig),
            'work_labeled_resource_chart': self.schedule_vis.work_labeled_resource_chart(VisualizationMode.ReturnFig),
            'gant_chart': self.schedule_vis.gant_chart(VisualizationMode.ReturnFig),
            'wg': self.wg_vis.work_graph_chart(VisualizationMode.ReturnFig)
        }



import json
from abc import ABC

import pandas as pd


def get_task_name_unique_mapping(path: str) -> 'NameMapper':
    """
    Gets mapping of our task names to the unique names
    :param path: path to the csv file
    :return: dict {our_name: unique_name}
    """
    df = read_tasks_df(path)
    return DictNameMapper({r[0]: r[1] for r in df.loc[:, ['task_name', 'unique_task_name']].to_numpy()})


def get_inverse_task_name_mapping(path: str) -> 'NameMapper':
    """
    Gets mapping of the unique names to our task names
    :param path: path to the csv file
    :return: dict {unique_name: our_name}
    """
    df = read_tasks_df(path)
    return DictNameMapper({r[1]: r[0] for r in df.loc[:, ['task_name', 'unique_task_name']].to_numpy()})


def read_tasks_df(path: str) -> pd.DataFrame:
    """
    Reads DataFrame with tasks
    :param path: path to the csv file
    :return: The DataFrame read
    """
    return pd.read_csv(path, sep=';', header=0)


def read_json(path: str) -> 'NameMapper':
    """
    Gets mapping of the unique names to our task names
    :param path: path to the .json file
    :return: NameMapper: our_name -> unique_name
    """
    with open(path) as f:
        mapper: dict[str, str] = json.load(f)
    return DictNameMapper({name[0]: name[1] for name in mapper.items()})


class NameMapper(ABC):
    def __getitem__(self, key):
        ...

    def get(self, key):
        ...

    def __setitem__(self, key, value):
        raise Exception('Trying to set a value to the NameMapper')


class DummyNameMapper(NameMapper):
    def __getitem__(self, item):
        return item

    def get(self, item):
        self.__getitem__(item)

    @property
    def _source(self):
        return self


class DictNameMapper(NameMapper):
    def __init__(self, source: dict[str, str]):
        self._source = source

    def get(self, item):
        self.__getitem__(item)

    def __getitem__(self, item):
        return self._source[item] if item in self._source else item


class ModelNameMapper(NameMapper):
    """
    NameMapper for Kovalchuk's model integration
    """

    # TODO: implement
    def __init__(self):
        self.__nie()

    def get(self, item):
        self.__getitem__(item)

    def __getitem__(self, item):
        self.__nie()

    def __nie(self):
        raise NotImplementedError('ModelNameMapper is not implemented')



from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.general_timeline import GeneralTimeline
from sampo.scheduler.timeline.just_in_time_timeline import JustInTimeTimeline
from sampo.scheduler.timeline.material_timeline import SupplyTimeline
from sampo.scheduler.timeline.momentum_timeline import MomentumTimeline
from sampo.scheduler.timeline.zone_timeline import ZoneTimeline



import os
import pathlib
from distutils.core import setup, Extension
from distutils.errors import DistutilsPlatformError, CCompilerError, DistutilsExecError

import numpy
from setuptools.command.build_ext import build_ext as build_ext_orig

ext_modules = [
    Extension("native",
              include_dirs=[numpy.get_include(), "timeEstimatorLibrary", "timeEstimatorLibrary/Windows"],
              sources=[
                       # "basic_types.h",
                       # "contractor.h",
                       # "dtime.h",
                       "dtime.cpp",
                       # "native.h",
                       "native.cpp",
                       # "pycodec.h",
                       # "python_deserializer.h",
                       "python_deserializer.cpp",
                       "chromosome_evaluator.cpp",
                       # "timeEstimatorLibrary/Windows/DLLoader.h",
                       # "workgraph.h"
              ],
              extra_compile_args=['-fopenmp', '/openmp'],
              extra_link_args=['-lgomp']),
]


class BuildFailed(Exception):
    pass

class CMakeExtension(Extension):

    def __init__(self, name):
        # don't invoke the original build_ext for this special extension
        super().__init__(name, sources=[])


class build_ext(build_ext_orig):

    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)
        super().run()

    def build_cmake(self, ext):
        cwd = pathlib.Path().absolute()

        # these dirs will be created in build_py, so if you don't have
        # any python sources to bundle, the dirs will be missing
        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)
        extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
        extdir.mkdir(parents=True, exist_ok=True)

        # example of cmake args
        config = 'Release'  # 'Debug' if self.debug else 'Release'
        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + str(extdir.parent.absolute()),
            '-DCMAKE_BUILD_TYPE=' + config,
            '-DCMAKE_C_COMPILER=gcc',
            '-DCMAKE_CXX_COMPILER=g++'
        ]

        # example of build args
        build_args = [
            '--config', config
            # '--', '-j4'
        ]

        os.chdir(str(build_temp))
        self.spawn(['cmake', str(cwd)] + cmake_args)
        if not self.dry_run:
            self.spawn(['cmake', '--build', '-DCMAKE_CXX_COMPILER=g++', '.'] + build_args)
        # Troubleshooting: if fail on line above then delete all possible
        # temporary CMake files including "CMakeCache.txt" in top level dir.
        os.chdir(str(cwd))


class ExtBuilder(build_ext):

    def run(self):
        try:
            build_ext.run(self)
        except (DistutilsPlatformError, FileNotFoundError):
            raise BuildFailed('File not found. Could not compile C extension.')

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError, ValueError):
            raise BuildFailed('Could not compile C extension.')


def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """
    setup_kwargs.update(
        {"ext_modules": [CMakeExtension(".")], "cmdclass": {"build_ext": build_ext}}
    )

setup(
    name='native',
    version='0.0.1',
    author='user',
    author_email='user@user.ru',
    description='C extension',
    ext_modules=ext_modules,
)
